
import numpy as np
from scipy.optimize import bisect
from pymatgen.electronic_structure.dos import FermiDos
import matplotlib
import matplotlib.pyplot as plt
from pynter.defects.pmg.pmg_dos import FermiDosCarriersInfo
from pynter.tools.utils import get_object_feature, select_objects, sort_objects
from monty.json import MontyDecoder, MSONable
import pandas as pd
import os.path as op
import json


class DefectsAnalysis:
    
    def __init__(self, entries, vbm, band_gap, sort_entries=True, occupation_function='MB'):
        """ 
        Class to compute defect properties starting from single calculations of defects.
        Args:
            entries: (list) 
                A list of DefectEntry objects.
            vbm: (float) 
                Valence Band energy to use for all defect entries.
                NOTE if using band shifting-type correction then this VBM
                should still be that of the GGA calculation
                (the bandedgeshifting_correction accounts for shift's
                contribution to formation energy).
            band_gap : (float)
                Band gap to use for all defect entries.
                NOTE if using band shifting-type correction then this gap
                should still be that of the Hybrid calculation you are shifting to. 
            occupation_function : (str) 
                Function to compute defect occupation. 
                "FD" for Fermi-Dirac, "MB" for Maxwell-Boltzmann.
        """
        self.entries = self.sort_entries(inplace=False,entries=entries) if sort_entries else entries
        self.vbm = vbm
        self.band_gap = band_gap
        self.occupation_function = occupation_function
        self.groups = self._group_entries()
        self.names = list(self.groups.keys())
        

    def __str__(self):     
        return self.get_dataframe().__str__()
            
    def __repr__(self):
        return self.__str__()
    
    def __iter__(self):
        return self.entries.__iter__()
    
    def _group_entries(self):
        groups = {}
        for e in self.entries:
            if e.name not in groups.keys():
                groups[e.name] = [e]
            else:
                groups[e.name].append(e)
        return groups
            
    
    def as_dict(self):
        """
        Returns:
            Json-serializable dict representation of DefectsAnalysis
        """
        d = {
        "@module": self.__class__.__module__,
        "@class": self.__class__.__name__,
        "entries": [entry.as_dict() for entry in self.entries],
        "vbm":self.vbm,
        "band_gap":self.band_gap
            }
        return d


    @classmethod
    def from_dict(cls,d):
        """
        Reconstitute a DefectsAnalysis object from a dict representation created using
        as_dict().

        Args:
            d (dict): dict representation of DefectsAnalysis.

        Returns:
            DefectsAnalysis object
        """
        entries = [MontyDecoder().process_decoded(entry_dict) for entry_dict in d['entries']]        
        vbm = d['vbm']
        band_gap = d['band_gap']
        return cls(entries,vbm,band_gap)
    
    
    @staticmethod
    def from_json(path_or_string):
        """
        Build Dataset object from json file or string.

        Parameters
        ----------
        path_or_string : (str)
            If an existing path to a file is given the object is constructed reading the json file.
            Otherwise it will be read as a string.

        Returns
        -------
        DefectsAnalysis object.

        """
        if op.isfile(path_or_string):
            with open(path_or_string) as file:
                d = json.load(file)
        else:
            d = json.loads(path_or_string)
        return DefectsAnalysis.from_dict(d)


    @property
    def elements(self):
        """
        List of all the elements involved in exchange of atoms with a reservoir 
        (elements present in entry.delta_atoms for all entries).
        """
        elements = []
        for entry in self.entries:
            for el in entry.delta_atoms.keys():
                if el not in elements:
                    elements.append(el)
        return elements


    def to_json(self,path='',**kwargs):
        """
        Save DefectsAnalysis object as json string or file.
        Parameters
        ----------
        path : (str), optional
            Path to the destination file. If '' the path is set to "self.path/self.name.json".
            If None a string is exported. 
        Returns
        -------
        d : (str)
            If path is not set a string is returned.
        """
        d = self.as_dict(**kwargs)
        if path == '':
            path = op.join(self.path,self.name+'.json')
        if path:
            with open(path,'w') as file:
                json.dump(d,file)
            return
        else:
            return d.__str__() 
        

    def binding_energy(self,name,fermi_level=0):
        """
        Args:
            name (string): name of defect complex as assigned in defect entry object
            fermi_level (float): Fermi level at which binding energy needs to be computed
        Returns:
            binding_energy (float)
        """
        stable_charges = self.stable_charges(None,fermi_level)
        charge = stable_charges[name][0]
        binding_energy = stable_charges[name][1]
        entry = self.select_entries(names=[name],charge=charge)[0]
        for sd in entry.defect.defects: #subtract energies of single defects
            dtype = sd.defect_type
            dspecie = sd.name.dspecie
            dname = self.select_entries(types=[dtype],defect_specie=dspecie)[0].name
            binding_energy = binding_energy - stable_charges[dname][1]    
        
        return binding_energy


    def carrier_concentrations(self,bulk_dos,temperature=300,fermi_level=0.):
        """
        Get intrinsic carrier concentrations by integrating over the density of states
        given a fixed Fermi level
        Args:
            bulk_dos: bulk system dos (pymatgen Dos object)
            temperature: Temperature to equilibrate fermi energies for
            fermi_level: (float) is fermi level relative to valence band maximum
                Default efermi = 0 = VBM energy         
        Returns:
            h,n in absolute values
        """
        
        fdos = FermiDosCarriersInfo(bulk_dos, bandgap=self.band_gap)
        _,fdos_vbm = fdos.get_cbm_vbm()    
        
        h, n = fdos.get_doping(fermi_level=fermi_level + fdos_vbm, temperature=temperature,carriers_values=True)
        
        return abs(h) , abs(n)


    def charge_transition_levels(self, energy_range=None,entries=None,get_integers=True):
        """
        Computes charge transition levels for all defect entries
        Args:
            energy_range (Tuple): Energy range to evaluate the charge transition levels in
                                  default to (-0.5, Eg + 0.5)    
            entries:
                List of entries to calculate. If None all entries are considered.
            get_integers : (bool)
                Save charges as integers. More convenient for plotting.
        Returns:
            Dictionary with defect name and list of tuples for charge transition levels:
                {name:[(q1,q2,ctl),(q2,q3,ctl),...}
        """                
        entries = entries if entries else self.entries
        if energy_range == None:
            energy_range = (-0.5,self.band_gap +0.5) 
        # creating energy array
        npoints = 3000
        step = abs(energy_range[1]-energy_range[0])/npoints
        e = np.arange(energy_range[0],energy_range[1],step)
        
        charge_transition_levels = {}
        # starting point is first E value
        previous_stable_charges = self.stable_charges(None,fermi_level=energy_range[0],entries=entries)
        for i in range(0,len(e)):
            stable_charges = self.stable_charges(None, fermi_level = e[i],entries=entries)
            for name in stable_charges:
                new_charge = stable_charges[name][0]
                previous_charge = previous_stable_charges[name][0]
                if new_charge != previous_charge:
                    if name not in charge_transition_levels.keys():
                        charge_transition_levels[name] = []
                    previous_stable_charges = stable_charges
                    if get_integers:
                        previous_charge = int(previous_charge)
                        new_charge = int(new_charge)
                    charge_transition_levels[name].append((previous_charge,new_charge,e[i]))
                            
        return charge_transition_levels        
    

    def defect_concentrations(self, chemical_potentials, temperature=300, fermi_level=0.,
                              fixed_concentrations=None,per_unit_volume=True):
        """
        Give list of all concentrations at specified efermi.
        If fixed_concentrations is provided the concentration of defect entries are 
        corrected according to the fixed provided values. More details can be found in 
        https://doi.org/10.1103/PhysRevB.106.134101 .
            
        Parameters
        ----------
        chemical_potentials: (dict)
            Dictionary of chemical potentials ({Element: number})   
        temperature: (float) 
            Temperature to produce concentrations at.
        fermi_level: (float) 
            Fermi level relative to valence band maximum. The default is 0. 
        fixed_concentrations: (dict)
            Dictionary with fixed concentrations. Keys can be simple element strings
            (or vacancies of elements in the format 'Vac_{el}') if that element needs to be 
            fixed across all defect species, alternatively defect entry names can be used as well 
            to target specific defect entries. The values are the concentrations.
        per_unit_volume: (bool)
            Get concentrations in cm^-3. If False they are per unit cell. The default is True.
        
        Returns:
        --------
            list of dictionaries of defect concentrations
        """
        concentrations = []
        if fixed_concentrations:
            dc = self.defect_concentrations(chemical_potentials,temperature,fermi_level,
                                            fixed_concentrations=None,per_unit_volume=per_unit_volume)
            frozen = fixed_concentrations 

        for e in self.entries:
            nsites = e.defect.site_concentration_in_cm3 if per_unit_volume else e.multiplicity
            # frozen defects approach
            if fixed_concentrations:
                c = e.defect_concentration(self.vbm, chemical_potentials,temperature,fermi_level,per_unit_volume,self.occupation_function)
                corr = self._get_frozen_correction(e,frozen,dc)
                c = c * corr
                defconc = SingleDefConc(name=e.name,charge=e.charge,conc=c,stable=bool(c<=nsites))
                concentrations.append(defconc)     
            
            else:
                c = e.defect_concentration(self.vbm, chemical_potentials,temperature,fermi_level,per_unit_volume,self.occupation_function)
                defconc = SingleDefConc(name=e.name,charge=e.charge,conc=c,stable=bool(c<=nsites))
                concentrations.append(defconc)
            

        return DefectConcentrations(concentrations)


    def _get_frozen_correction(self,e,frozen,dc):
        corr = 1
        for dn in e.name:
            typ, specie, name = dn.dtype, dn.dspecie, dn.name
            if typ == 'Vacancy':
                k = name
                if k in frozen.keys():
                    eltot = dc.get_element_total(specie,vacancy=True)
                    corr = corr * (frozen[k]/eltot)
            else:
                if name in frozen.keys():
                    dtot = 0
                    for n,c in dc.total.items(): # sum over all defects containing the specific specie
                        if name in n:
                            dtot += c
                    corr = corr * (frozen[name]/dtot) 
                else:
                    k = specie
                    if k in frozen.keys():
                        eltot = dc.get_element_total(specie,vacancy=False)
                        corr = corr * (frozen[k]/eltot)
        return corr

    
    def filter_entries(self,inplace=False,entries=None,mode='and',exclude=False,types=None,
                       elements=None,names=None,function=None,**kwargs):
        """
        Filter entries based on different criteria. Return another DefectsAnalysis object.

        Parameters
        ----------
        inplace : (bool)
            Apply changes to current DefectsAnalysis object.
        entries : (list)
            List of defect entries.
        mode : (str), optional
            Filtering mode, possibilities are: 'and' and 'or'. The default is 'and'. 
        exclude : (bool), optional
            Exclude the entries satisfying the criteria instead of selecting them. The default is False.
        types : (list), optional
            Class name of the defect in the entry.
        elements : (list), optional
            List of symbols of elements that need to belong to the defect specie.
            If None this criterion is ignored. The default is None.
        names : (list)
            List of entry names.
        function : (function), optional
            Specific funtion for more complex criteria. The function must take a DefectEntry
            object as argument and return a bool.
        **kwargs : (dict)
            Properties that the jobs need to satisfy. Keys are referred to attributes/methods 
            of the defect entry. To address more than one condition relative to
            the same attribute, use lists or tuples (e.g. charge=[0,1]).

        Returns
        -------
        DefectsAnalysis object.
        """
        output_entries = self.select_entries(entries=entries,mode=mode,exclude=exclude,types=types,
                                             elements=elements,names=names,function=function,**kwargs)
        
        if inplace:
            self.entries = output_entries
            return
        else:
            return DefectsAnalysis(output_entries, self.vbm, self.band_gap)


    def formation_energies(self,chemical_potentials,fermi_level=0,entries=None):
        """
        Creating a dictionary with names of single defect entry as keys and
        a list of tuples (charge,formation_energy) as values
        Every defect name identifies a type of defect
        Args:
            chemical_potentials:
                a dictionnary of {Element:value} giving the chemical
                potential of each element
            fermi_level:
                Value of fermi level to use for calculating formation energies
            entries:
                List of entries to calculate. If None all entries are considered.
        Returns:
            {name: [(charge,formation energy)] }
        """
        formation_energies = {}
        entries = entries if entries else self.entries
        for entry in entries:
            name = entry.name
            charge = entry.charge
            eform = entry.formation_energy(self.vbm,chemical_potentials,fermi_level=fermi_level)
            if name in formation_energies:
                formation_energies[name].append((charge,eform))
            else:
                formation_energies[name] = []
                formation_energies[name].append((charge,eform))
        
        return formation_energies
        
           
    def get_charge_transition_level(self,name,q1,q2):
        """
        Args:
            chemical_potentials: 
                a dictionnary of {Element:value} giving the chemical
                potential of each element
            name: (str) Name of the defect system to get the charge transition levels for
            q1, q2 (int) : Charges for whcih you want to evaluate the transition levels
        Returns: Charge transition level (float)
        """
        chemical_potentials = None
        formation_energies = self.formation_energies(chemical_potentials)
        
        for d in formation_energies[name]:
            if d[0] == q1:
                E1 = d[1]
            if d[0] == q2:
                E2 = d[1]
        charge_transition_level = -1*(E1 - E2)/(q1 - q2)
        
        return charge_transition_level
        

    def get_dataframe(self,entries=None,pretty=False,include_bulk=False,display=[]):
        """
        Get DataFrame to display entries. 

        Parameters
        ----------
        entries : (list), optional
            Entries to display. If None all entries are displayed. The default is None.
        pretty : (bool), optional
            Optimize DataFrame for prettier visualization.
        include_bulk: (bool), optional
            Include bulk composition and space group for each entry in DataFrame.
        display: (list)
            List of strings with defect entry attributes or method results to display.

        Returns
        -------
        df : 
            DataFrame object.
        """
        if not entries:
            entries = self.entries
        d = {}
        index = []
        table = []
        for e in entries:
            symbol = e.symbol
            if pretty:
                index.append(symbol)
            else:
                index.append(e.name)
            d = {}
            if include_bulk:
                d['bulk composition'] = e.bulk_structure.composition.formula
                d['bulk space group'] = e.bulk_structure.get_space_group_info()
            if not pretty:
                d['symbol'] = symbol    
                d['delta atoms'] = e.delta_atoms
            d['charge'] = e.charge
            d['multiplicity'] = e.multiplicity
            if display:
                for feature in display:
                    if isinstance(feature,list):
                        key = feature[0]
                        for k in feature[1:]:
                            key = key + '["%s"]'%k
                    else:
                        key = feature
                    d[key] = get_object_feature(e,feature)
            table.append(d)
        df = pd.DataFrame(table,index=index)
        if pretty:
            df.index.name = 'symbol'
        else:
            df.index.name = 'name'
        return df
    
                
    def plot(self,chemical_potentials,entries=None,xlim=None,ylim=None,title=None,
             fermi_level=None,plotsize=1,fontsize=1.2,show_legend=True,
             format_legend=True,get_subplot=False,subplot_settings=None):
        """
        Produce defect Formation energy vs Fermi energy plot
        Args:
            chemical_potentials:
                Chempots object containing the chemical
                potential of each element.
            entries: 
                List of entries to calculate. If None all entries are considered.
            xlim:
                Tuple (min,max) giving the range of the x (fermi energy) axis.
            ylim:
                Tuple (min,max) giving the range for the formation energy axis.
            fermi_level:
                float to plot Fermi energy position.
            plotsize:
                float or tuple. Can be float or tuple for different x and y sizes
                multiplier to change plotsize.
            fontsize:
                float  multiplier to change fontsize.
            show_legend:
                Bool for showing legend.
            format_legend:
                Bool for getting latex-like legend based on the name of defect entries.
            get_subplot:
                Bool for getting subplot.
            subplot_settings:
                List with integers for subplot setting on matplotlib (plt.subplot(nrows,ncolumns,index)). 
        Returns:
            matplotlib object
        """
        entries = entries if entries else self.entries
        matplotlib.rcParams.update({'font.size': 10*fontsize})        
        formation_energies = self.formation_energies(chemical_potentials,fermi_level=0,entries=entries)
        if xlim == None:
            xlim = (-0.5,self.band_gap+0.5)        
        npoints = 200
        step = abs(xlim[1]+0.1-xlim[0])/npoints
        x = np.arange(xlim[0],xlim[1]+0.1,step)
        try:
            if len(plotsize)==2:
                pass
        except:
            plotsize = (plotsize,plotsize)
        
        if get_subplot:
            if subplot_settings[2] == 1:
                plt.figure(figsize=(8*plotsize[0],6*plotsize[1]))               
            plt.subplot(subplot_settings[0],subplot_settings[1],subplot_settings[2])
            plt.grid()
        else:
            plt.figure(figsize=(8*plotsize[0],6*plotsize[1]))
            plt.grid()
            
        for name in formation_energies:
            energy = np.zeros(len(x))
            emin = np.zeros(len(x))
            x_star = []
            y_star = []
            q_previous = None
            # this is faster than using self.stable_charges
            for i in range(0,npoints):
                emin[i] = 1e40
                for d in formation_energies[name]:
                    q = d[0]
                    e0 = d[1]
                    energy[i] = e0 + q*x[i]
                    # finding most stable charge state
                    if energy[i] < emin[i]:
                        emin[i] = energy[i]
                        q_stable = q
                # getting data to plot transition levels        
                if q_stable != q_previous:
                    if q_previous != None:
                        x_star.append(x[i])
                        y_star.append(emin[i])
                    q_previous = q_stable       

            if format_legend:
                label_txt = name.symbol
            else:
                label_txt = name            

            plt.plot(x,emin,label=label_txt,linewidth=3)
            plt.scatter(x_star,y_star,s=120,marker='*')
                        
        plt.axvline(x=0.0, linestyle='-', color='k', linewidth=2)  # black dashed lines for gap edges
        plt.axvline(x=self.band_gap, linestyle='-', color='k',
                    linewidth=2)        
        if fermi_level:
            plt.axvline(x=fermi_level, linestyle='dashed', color='k', linewidth=1.5, label='$\mu _{e}$')                
        # shaded areas
        plt.axvspan(xlim[0], 0, facecolor='k', alpha=0.2)
        plt.axvspan(self.band_gap, xlim[1]+0.1, facecolor='k', alpha=0.2)
        plt.hlines(0,xlim[0],xlim[1]+0.1,colors='k',linestyles='dashed',alpha=0.5)
        plt.xlim(xlim)
        if ylim: 
            plt.ylim(ylim) 
        plt.xlabel('Fermi level (eV)')
        plt.ylabel('Formation energy (eV)')
        if title:
            plt.title(title)
        if show_legend:    
            plt.legend()
     #   plt.grid()

        return plt
     
        
    def plot_binding_energies(self, names=None, xlim=None, ylim=None, size=1,format_legend=True):
        """
        Plot binding energies for complex of defects as a function of the fermi level
        Args:
            names: 
                List of strings with names of DefectEntry. If None all DefectEntry
                objects are plotted.
            xlim:
                Tuple (min,max) giving the range of the x (fermi energy) axis
            ylim:
                Tuple (min,max) giving the range for the formation energy axis
            size:
                Float multiplier to change plot size
            format_legend:
                Bool for getting latex-like legend based on the name of defect entries
        Returns:
            matplotlib object
        """        
        plt.figure(figsize=(8*size,6*size))
        matplotlib.rcParams.update({'font.size': 10*1.8*size}) 
        if xlim==None:
            xlim = (-0.5,self.band_gap+0.5)
        # building array for x values (fermi level)    
        ef = np.arange(xlim[0],xlim[1]+0.1,(xlim[1]-xlim[0])/200)        
        binding_energy = np.zeros(len(ef))        
        if not names:
            names = []
            for e in self.entries:
                if e.defect_type == 'DefectComplex':
                    if e.name not in names:
                        names.append(e.name)        
        # getting binding energy at different fermi levels for every name in list
        for name in names:
            label = self.select_entries(names=[name])[0].symbol if format_legend else name
            for i in range(0,len(ef)):
                binding_energy[i] = self.binding_energy(name,fermi_level=ef[i])            
            plt.plot(ef,binding_energy, linewidth=2.5*size,label=label)
            
        plt.axvline(x=0.0, linestyle='-', color='k', linewidth=2)  # black lines for gap edges
        plt.axvline(x=self.band_gap, linestyle='-', color='k',
                    linewidth=2)        
        # shaded areas
        plt.axvspan(xlim[0], 0, facecolor='k', alpha=0.2)
        plt.axvspan(self.band_gap, xlim[1], facecolor='k', alpha=0.2)
        plt.hlines(0,xlim[0],xlim[1],colors='k',linestyles='dashed',alpha=0.5)
        plt.legend()
        plt.xlim(xlim)
        if ylim: 
            plt.ylim(ylim) 
        plt.xlabel('Fermi level (eV)')
        plt.ylabel('Binding energy (eV)')
        
        return plt
    
    
    def plot_ctl(self, entries=None,ylim=None, size=1, fermi_level=None, format_legend=True,get_integers=True):
        """
        Plotter for the charge transition levels
        Args:
            entries: List of entries to calculate. If None all entries are considered.
            ylim (Tuple): y-axis limits.
            fermi_level (float) : float to plot Fermi energy position.
            size (float) : Float multiplier for plot size.
            format_legend (bool): Bool for getting latex-like legend based on the name of defect entries .
            get_integers (bool): Get CTLs as integers.
        Returns:
            matplotlib object    
        """        
        plt.figure(figsize=(10*size,10*size))         
        if ylim == None:
            ylim = (-0.5,self.band_gap +0.5)        
        charge_transition_levels = self.charge_transition_levels(entries=entries,get_integers=get_integers)
        number_defects = len(charge_transition_levels)   
        x_max = 10
        interval = x_max/(number_defects + 1)
        x = np.arange(0,x_max,interval)
        # position of x labels
        x_ticks_positions = []
        for i in range(0,len(x)-1):
            x_ticks_positions.append((x[i+1]-x[i])/2 + x[i])            
        x_ticks_labels = []
        for name in charge_transition_levels:
            x_ticks_labels.append(name)        
        # draw vertical lines to separte defect types
        for i in x:
            plt.axvline(x=i, linestyle='-', color='k', linewidth=1.2, alpha=1, zorder=1)
        xlim = (x[0],x[-1])        
        #VBM and CBM shaded
        plt.axhspan(ylim[0], 0, facecolor='grey', alpha=0.9, zorder=2)
        plt.axhspan(self.band_gap,ylim[1], facecolor = 'grey', alpha=0.9, zorder=2)                
        # plot CTL
        for i in range(0,len(x_ticks_labels)):
            name = x_ticks_labels[i]
            for ctl in charge_transition_levels[name]:
                energy = ctl[2]
                plt.hlines(energy,x[i],x[i+1],colors='k',linewidth=2.25, zorder=3)
                charge1 = '+' + str(ctl[1]) if ctl[1] > 0 else str(ctl[1])
                charge2 = '+' + str(ctl[0]) if ctl[0] > 0 else str(ctl[0])
                label_charge = '(' + charge2 + '/' + charge1 + ')'
                font_space = abs(ylim[1]-ylim[0]) / 100
                if energy < ylim[1] and energy > ylim[0]:
                    plt.text(x[i]+(interval/2)*2/number_defects ,energy+font_space,label_charge,fontsize=16*size)        
        # format latex-like legend
        if format_legend:    
             for name in x_ticks_labels:            
                x_ticks_labels[x_ticks_labels.index(name)] = name.symbol               
        if fermi_level:
            plt.axhline(y=fermi_level, linestyle='dashed', color='k', linewidth=1.5, label='$\mu _{e}$')   
        
        plt.text(x[-1]+interval/8,-0.3,'VB',fontsize=25*size)
        plt.text(x[-1]+interval/8,self.band_gap+0.2,'CB',fontsize=25*size)
        plt.xticks(ticks=x_ticks_positions,labels=x_ticks_labels,fontsize = (25-number_defects)*size)
        plt.tick_params(axis='x',length=0,width=0)
        plt.yticks(fontsize=16*size)
        plt.xlim(xlim)  
        plt.ylim(ylim)
        plt.ylabel('Energy(eV)',fontsize=20*size)  
        
        return plt    


    def select_entries(self,entries=None,mode='and',exclude=False,types=None,elements=None,
                       names=None,function=None,**kwargs):
        """
        Find entries based on different criteria. Returns a list of DefectEntry objects.

        Parameters
        ----------
        entries : (list)
            List of defect entries.
        mode : (str), optional
            Filtering mode, possibilities are: 'and' and 'or'. The default is 'and'. 
        exclude : (bool), optional
            Exclude the entries satisfying the criteria instead of selecting them. The default is False.
        types : (list), optional
            Class name of the defect in the entry.
        elements : (list), optional
            List of symbols of elements that need to belong to the defect specie.
            If None this criterion is ignored. The default is None.
        names : (list)
            List of entry names.
        function : (function), optional
            Specific funtion for more complex criteria. The function must take a DefectEntry
            object as argument and return a bool.
        **kwargs : (dict)
            Properties that the jobs need to satisfy. Keys are referred to attributes/methods 
            of the defect entry. To address more than one condition relative to
            the same attribute, use lists or tuples (e.g. charge=[0,1]).

        Returns
        -------
        List of DefectEntry objects.
        """        
        input_entries = entries if entries else self.entries 
        functions = []
        
        if types:
            def ftypes(entry):
                return entry.defect_type in types
            functions.append(ftypes)
        
        if elements:
            def felements(entry):
                for name in entry.name:
                    if name.dspecie in elements:
                        return True
            functions.append(felements)
                
        if names:
            def fnames(entry):
                return entry.name in names
            functions.append(fnames)
        
        if function:
            functions.append(function)
            
        return select_objects(objects=input_entries,mode=mode,exclude=exclude,
                              functions=functions,**kwargs)
    
    
    def solve_fermi_level(self,chemical_potentials,bulk_dos,temperature=300,
                                    fixed_concentrations=None,external_defects=[],xtol=1e-05):
        """
        Solve charge neutrality and get the value of Fermi level at thermodynamic equilibrium.
        
        Parameters
        ----------
        chemical_potentials : (Dict)
            Dictionary of chemical potentials in the format {Element('el'):chempot}.
        bulk_dos : (CompleteDos object)
            Pymatgen CompleteDos object of the DOS of the bulk system.
        temperature : (float), optional
            Temperature to equilibrate the system to. The default is 300.
        fixed_concentrations: (dict)
            Dictionary with fixed concentrations. Keys are defect entry names in the standard
            format, values are the concentrations (ex {'Vac_A':1e20}) . For more info, 
            read the documentation of the defect_concentrations method.
        external_defects : (list)
            List of SingleDefConc objects containing external defect concentrations 
            (not present in defect entries).
        xtol: Tolerance for bisect (scipy) to solve charge neutrality. The default is 1e-05.

        Returns
        -------
        (float)
            Fermi level dictated by charge neutrality.
        """
        
        fdos = FermiDosCarriersInfo(bulk_dos, bandgap=self.band_gap)
        _,fdos_vbm = fdos.get_cbm_vbm()

        def _get_total_q(ef):
            qd_tot = sum([
                d.charge * d.conc
                for d in self.defect_concentrations(chemical_potentials,temperature,ef,
                                                    fixed_concentrations)])
            for d_ext in external_defects:
                qd_tot += d_ext.charge * d_ext.conc                
            qd_tot += fdos.get_doping(fermi_level=ef + fdos_vbm, temperature=temperature)
            return qd_tot
                       
        return bisect(_get_total_q, -1., self.band_gap + 1.,xtol=xtol)
    
    
    def sort_entries(self,inplace=False,entries=None,features=['name','charge'],reverse=False):
        """
        Sort defect entries with different criteria.

        Parameters
        ----------
        inplace : (bool), optional
            Reset the self.entries attibute with sorted entries. Only works if entries input
            is not given. The default is False.
        entries : (list), optional
            List of defect entries to sort. If None self.entries is used. The default is None.
        features : (list), optional
            List of strings with attribute/method names. The default is ['name','charge'].
        reverse : (bool)
            Reverse order.

        Returns
        -------
        (list)
            List of defect entries.
        """
        inplace = False if entries else inplace
        entries = entries if entries else self.entries
        sorted_entries = sort_objects(objects=entries, features=features)
        if inplace:
            self.entries = sorted_entries
        else:
            return sorted_entries
            
    
    def stable_charges(self,chemical_potentials,fermi_level=0,entries=None):
        """
        Creating a dictionary with names of single defect entry as keys and
        as value a tuple (charge,formation_energy) that gives the most stable 
        charge state at the inserted fermi level.
        Every defect name identifies a type of defect
        Args:
            chemical_potentials:
                a dictionnary of {Element:value} giving the chemical
                potential of each element
            fermi_level:
                Value of fermi level to use for calculating formation energies 
            entries:
                List of entries to calculate. If None all entries are considered.
        Returns:
            {name:(stable charge, formation energy)}
       """
        formation_energies = self.formation_energies(chemical_potentials,fermi_level,entries)        
        stable_charges = {}
        for name in formation_energies:
            emin = 1e40
            for d in formation_energies[name]:
                q = d[0]
                energy = d[1]
                # finding most stable charge state
                if energy < emin:
                    emin = energy
                    q_stable = q
            stable_charges[name] = (q_stable,emin)
            
        return stable_charges
        

    def _get_total_charge(self,fermi_level,chemical_potentials, bulk_dos, temperature=300, 
                          fixed_concentrations=None,external_defects=[], per_unit_volume=True):
        fdos = FermiDos(bulk_dos, bandgap=self.band_gap)
        _,fdos_vbm = fdos.get_cbm_vbm()    
        qd_tot = sum([
            d.charge * d.conc
            for d in self.defect_concentrations(chemical_potentials,temperature,fermi_level,
                                                fixed_concentrations,per_unit_volume)])
        for d_ext in external_defects:
            qd_tot += d_ext.charge * d_ext.conc
        qd_tot += fdos.get_doping(fermi_level=fermi_level + fdos_vbm, temperature=temperature)
        return qd_tot



class SingleDefConc(MSONable):
    
    def __init__(self,name,charge,conc,stable=True):
        """
        Object to store defect concentrations data. It is also subscribtable like a dictionary.

        Parameters
        ----------
        name : (str)
            Name of defect specie.
        charge : (int)
            Charge of defect specie.
        conc : (float)
            Concentration value.
        stable : (bool)
            If formation energy is positive, i.e. the concentration < Nsites.
        """
        self.name = name
        self.charge = charge
        self.conc = conc
        self.stable = stable
        
    def __repr__(self):
        s = 'charge=%.1f, conc=%.2e, name=%s, stable=%s' %(self.charge,self.conc,self.name,self.stable)
        return s
    
    def __print__(self):
        return self.__repr__()
    
    def __getitem__(self,key):
        return getattr(self,key)
        
        
class DefectConcentrations:
    
    def __init__(self,concentrations):
        """
        Class to store sets of defect concentrations (output of concentration calculations with DefectsAnalysis).
        List of SingleDefConc objects. Subscriptable like a list.

        Parameters
        ----------
        concentrations : (list)
            List of SingleDefConc objects.
        """
        self.concentrations = concentrations
        d = {}
        # calculate total concentrations
        for c in self.concentrations:
            gname = c.name # previously was the name without label
            if gname not in d.keys():
                d[gname] = 0
            d[gname] += c.conc    
        self._total = d
      #  store stable concentrations
        conc_stable = []
        for n in self.names:
            concs = self.select_concentrations(name=n)
            cmax = SingleDefConc(name='',conc=None,charge=0,stable=True) # dummy object
            for c in concs:
                if cmax.conc is None or c.conc >= cmax.conc:
                    cmax = c
            conc_stable.append(concs[concs.index(cmax)])
        self._stable = conc_stable
        
        
    def __len__(self):
        return len(self.concentrations)
        
    def __iter__(self):
        return self.concentrations.__iter__()
    
    def __getitem__(self,i):
        return self.concentrations[i]
    
    def __repr__(self):
        return self.__print__()

    def __print__(self):
        return '['+ '\n'.join([c.__print__() for c in self.concentrations])+']'  

   
    def as_dict(self):
        d = [c.as_dict() for c in self.concentrations]
        return d

    @classmethod 
    def from_dict(cls,d):
        dc = [SingleDefConc.from_dict(c) for c in d]
        return cls(dc)
 
    @property
    def elemental(self):
        """
        Dictionary with element (or element vacancy) as keys and total element concentration as values.
        """
        d = {}
        for c in self:
            for dn in c.name:
                if dn.dspecie not in d.keys():
                    if dn.dtype == 'Vacancy':
                        ekey = dn.name
                        d[ekey] = self.get_element_total(dn.dspecie,vacancy=True)
                    else:
                        ekey = dn.dspecie
                        d[ekey] = self.get_element_total(dn.dspecie,vacancy=False)
        return d
                    
    
    @property
    def elements(self):
        """
        List of element (or vacancies) symbols.
        """
        return self.elemental.keys()

                
    @property
    def names(self):
        """
        List of names of all defect species
        """
        return list(self.total.keys())

    @property
    def stable(self):
        """
        Get concentrations of only stable charge states. List of SingleDefConc objects.
        """
        return self._stable    

    @property
    def total(self):
        """
        Get total concentrations for every defect specie. Format is a dict ({'name':conc})
        """
        return self._total

    
    def filter_concentrations(self,inplace=False,mode='and',exclude=False,names=None,
                              charges=None,indexes=None,function=None,**kwargs):
        """
        Filter concentrations based on different criteria.

        Parameters
        ----------
        inplace : (bool), optional
            If True update the current object, otherwise returns a new DefectConcentrations object.
            The default is False.
        mode : (str), optional
            Selection mode. available are 'and','or'. The default is 'and'.
        exclude : (bool), optional
            Exclude the entries matching the criteria. The default is False.
        names : (list), optional
            Names of the defect.
        charges : (list), optional
            Charges of the defect.
        indexes : (list), optional
            Indexes of defect in the concentrations list.
        function : (function), optional
            Specific funtion for more complex criteria. The function must take a 
            SingleDefConc object as argument and return a bool.
        **kwargs : (dict)
            Criteria for selection. They need to be attributes of SingleDefConc.

        Returns
        -------
        output_concs : (list)
            Filtered DefectConcentrations object.
        """
        output_concs = self.select_concentrations(mode=mode,exclude=exclude,names=names,
                                                  charges=charges,indexes=indexes,
                                                  function=function,**kwargs)
                    
        if inplace:
            self.concentrations = output_concs
            return
        else:
            return DefectConcentrations(output_concs)
            
        
    def get_element_total(self,element,vacancy=False):
        """
        Get total concentration of every element (summed also across different species)

        Parameters
        ----------
        element : (str)
            Target element.
        vacancy : (bool), optional
            Weather the target element is a vacancy. The default is False.

        Returns
        -------
        eltot : (float)
            Total concentration of target element.
        """
        eltot = 0
        for c in self:
            for dn in c.name:
                if element == dn.dspecie:
                    if vacancy and dn.dtype=='Vacancy':
                        eltot += c.conc
                    elif vacancy == False:
                        eltot += c.conc
        return eltot
        

    def select_concentrations(self,concentrations=None,mode='and',exclude=False,names=None,
                    charges=None,indexes=None,function=None,**kwargs):
        """
        Select concentrations based on different criteria.

        Parameters
        ----------
        concentrations : (list), optional
            List of SingleDefConc to select from. If None self.concentrations is used.
            The default is None.
        mode : (str), optional
            Selection mode. available are 'and','or'. The default is 'and'.
        exclude : (bool), optional
            Exclude the entries matching the criteria. The default is False.
        names : (list), optional
            Names of the defect.
        charges : (list), optional
            Charges of the defect.
        indexes : (list), optional
            Indexes of defect in the concentrations list.
        function : (function), optional
            Specific funtion for more complex criteria. The function must take a 
            SingleDefConc object as argument and return a bool.
        **kwargs : (dict)
            Criteria for selection. They need to be attributes of SingleDefConc.

        Returns
        -------
        output_concs : (list)
            List of SingleDefConc objects.
        """
        input_concs = concentrations if concentrations else self.concentrations.copy()
        functions = []
        
        if names is not None:
            def fnames(c):
                return c.name in names
            functions.append(fnames)
        
        if charges is not None:
            def fcharges(c):
                return c.charge in charges
            functions.append(fcharges)
        
        if indexes is not None:
            def findexes(c):
                return input_concs.index(c) in indexes
            functions.append(findexes)
        
        if function is not None:
            functions.append(function)
            
        return select_objects(objects=input_concs,mode=mode,exclude=exclude,
                              functions=functions,**kwargs)
    
                    
        
        
        



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed May 10 16:13:48 2023

@author: villa
"""

from pymatgen.core.structure import Structure

def get_structure_Si():
    structure = Structure.from_dict(
    {'@module': 'pymatgen.core.structure',
     '@class': 'Structure',
     'charge': 0,
     'lattice': {'matrix': [[3.32548851, 0.0, 1.91997169],
       [1.10849617, 3.13530064, 1.91997169],
       [0.0, 0.0, 3.83994338]],
      'pbc': (True, True, True),
      'a': 3.839943374653261,
      'b': 3.839943378813096,
      'c': 3.83994338,
      'alpha': 59.99999998977525,
      'beta': 59.99999995393976,
      'gamma': 60.00000000512866,
      'volume': 40.036809671145996},
     'sites': [{'species': [{'element': 'Si', 'occu': 1}],
       'abc': [0.875, 0.875, 0.875],
       'xyz': [3.879736595, 2.74338806, 6.719900914999999],
       'label': 'Si',
       'properties': {}},
      {'species': [{'element': 'Si', 'occu': 1}],
       'abc': [0.125, 0.125, 0.125],
       'xyz': [0.554248085, 0.39191258, 0.959985845],
       'label': 'Si',
       'properties': {}}]}
        )
    return structure



def get_incar_settings():
    incar_settings = {'ALGO': 'Normal',
     'AMIX': 0.2,
     'EDIFF': 1e-06,
     'EDIFFG': -0.05,
     'ENCUT': 550,
     'IBRION': 2,
     'ISIF': 2,
     'ISMEAR': 0,
     'ISPIN': 1,
     'KPAR': 4,
     'LCHARG': '.TRUE.',
     'LORBIT': 10,
     'LREAL': '.FALSE.',
     'LWAVE': '.TRUE.',
     'NELM': 200,
     'NSW': 0,
     'SIGMA': 0.05,
     'SYSTEM': 'Si',
     '#### Default PBE: system': 'Si',
     'ISYM': 2}
    return incar_settings



def get_job_settings():
    job_settings = {
     'add_automation': 'automation_vasp.py --contcar --chgcar --wavecar --check-kpoints --error-check',
     'add_lines_body': None,
     'add_lines_header': None,
     'add_stop_array': True,
     'array_size': None,
     'cores_per_node': 24,
     'email': 'test@pynter.com',
     'error_filename': 'err.%j',
     'filename': 'job.sh',
     'memory_per_cpu': 2400,
     'modules': ['intel/2019.2', 'intel/2019.3', 'intelmpi/2019.3', 'fftw/3.3.8'],
     'name': 'Si-BS_PBE-el-str_3',
     'nodes': 1,
     'output_filename': 'out.%j',
     'partition': 'deflt',
     'path_exe': '/home/vasp-5-3-3',
     'processor': 'avx2',
     'project_id': 'project0000',
     'timelimit': '00:30:00'}
    return job_settings


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Apr  4 17:10:46 2023

@author: villa
"""

import warnings
from pymatgen.core.sites import PeriodicSite
from pymatgen.core.periodic_table import Element
import os.path as op
import os
from pynter.tools.structure import is_site_in_structure, is_site_in_structure_coords, sort_sites_to_ref_coords, write_extxyz_file
from pynter.defects.defects import Vacancy,Substitution,Interstitial,DefectComplex
from pymatgen.core.trajectory import Trajectory

"""Interstitial generator to be re-implemented using the new pymatgen defects"""
    
def create_interstitial_structures(structure,elements,supercell_size=None,**kwargs):
    """
    Create structures with interstitials based on Voronoi with pymatgen. 

    Parameters
    ----------
    structure : (Structure)
        Bulk structure.
    elements : (list)
        List of element symbols.
    supercell_size : (int), optional
        Input for the make_supercell function of the Structure class.
        If None the input structure is not modified. The default is None.
    kwargs: 
        Arguments to pass to VoronoiInterstitialGenerator:
            clustering_tol: Tolerance for clustering the Voronoi nodes.
            min_dist: Minimum distance between an interstitial and the nearest atom.
            ltol: Tolerance for lattice matching.
            stol: Tolerance for structure matching.
            angle_tol: Angle tolerance for structure matching.
            kwargs: Additional keyword arguments for the ``TopographyAnalyzer`` constructor.

    Returns
    -------
    interstitial_structures : (dict)
        Dictionary with element as keys and list of structures as values.
    """
    from pymatgen.analysis.defects.generators import VoronoiInterstitialGenerator
    bulk_structure = structure.copy()
    if supercell_size:
        bulk_structure.make_supercell(supercell_size)
    generator = VoronoiInterstitialGenerator().generate(bulk_structure,elements)
    interstitial_structures = {}
    for inter in generator:
        el = inter.site.specie.element.symbol
        if el not in interstitial_structures.keys():
            interstitial_structures[el] = []
        interstitial_structures[el].append(inter.defect_structure)
            
    return interstitial_structures
    

def create_substitution_structures(structure,replace,supercell_size=1):
    """
    Create structures with substitutions starting from a bulk structure (unit cell or supercell).

    Parameters
    ----------
    structure : Structure
        Bulk structure, both unit cell or supercell can be used as input.
    replace : (str), optional
        Dict with element symbol of specie to be replaced as keys and element 
        symbol of the species to be replaced with as values ({'old_El':{new_El}}).
    supercell_size : (int or numpy array), optional
        Input for the generate_defect_structure function of the old pymatgen Substitution class.

    Returns
    -------
    sub_structures : (dict)
        Dictionary with substitution types as keys and structures as values.
    """
    from pynter.defects.pmg.pmg_defects_core import Substitution
    sub_structures={}
    bulk_structure = structure.copy()
    
    for el_to_sub in replace:
        for el in bulk_structure.composition.elements:
            s = bulk_structure.copy()
            for site in s.sites:
                if el.symbol == el_to_sub:
                    if site.specie == el:
                        sub_site = site
                        sub_el = replace[el.symbol]
                        defect_site = PeriodicSite(sub_el,sub_site.frac_coords,sub_site.lattice)
                        structure = Substitution(s,defect_site).generate_defect_structure(supercell_size)
                        sub_structures[f'{sub_el}-on-{el.symbol}'] = structure
                        break

    return sub_structures


def create_vacancy_structures(structure,elements=None,supercell_size=None):
    """
    Create structures with vacancies starting from a bulk structure (unit cell or supercell).

    Parameters
    ----------
    structure : Structure
        Bulk structure, both unit cell or supercell can be used as input.
    elements : (str), optional
        Symbol of the elements for which vacancies are needed.
        If None all of the elements are considered. The default is None.
    supercell_size : (int or numpy array), optional
        Input for the make_supercell function of the Structure class.
        If None the input structure is not modified. The default is None.

    Returns
    -------
    vac_structures : (dict)
        Dictionary with vacancy types as keys and structures as values.

    """
    vac_structures={}
    bulk_structure = structure.copy()
    if supercell_size:
        bulk_structure.make_supercell(supercell_size)
    if not elements:
        elements = [el.symbol for el in bulk_structure.composition.elements]
    
    for el in bulk_structure.composition.elements:
        s = bulk_structure.copy()
        for site in s.sites:
            if el.symbol in elements:
                if site.specie == el:
                    s.remove_sites([bulk_structure.index(site)])
                    vac_structures[f'{el.symbol}'] = s
                    break

    return vac_structures

 
def create_def_structure_for_visualization(structure_defect,structure_bulk,defects=None,sort_to_bulk=False,tol=1e-03):
    """
    Create defect structure for visualization in OVITO. The vacancies are shown by inserting 
    in the vacant site the element of same row and next group on the periodic table.
    If sort_to_bulk is True the Sites are sorted to match the Bulk structure.

    Parameters
    ----------
    structure_defect : (Pymatgen Structure object)
        Defect structure.
    structure_bulk : (Pymatgen Structure object)
        Bulk structure.
    defects : (tuple or list). 
        List of defect objects.
    sort_to_bulk : (bool)
        Sort Sites of the defect structure to match the order of coordinates in the bulk structure
        (useful if the non-relaxed defect structure is not available). 
        If False only the dummy atoms are inserted and not further changes are made.
    tol: (float)
        Tolerance for site comparison. The distance between sites in target and reference stucture is used, 
        periodicity is accounted for. The tolerance is normalized with respect to lattice vector size. 
        The default is 1e-03.

    Returns
    -------
    new_structure (Pymatgen Structure object)
        Structure with dummy atoms as vacancies and interstitials in the bottom.
        The order of the Sites follow the order of the Bulk structure.

    """
    df = structure_defect.copy()
    bk = structure_bulk.copy()
    extra_sites=[]
    if defects:
        dfs = defects
    else:
        df_found = defect_finder(df,bk,tol=tol)
        if df_found.defect_type=='DefectComplex':
            dfs = df_found.defects
        else:
            dfs = [df_found]
   
    for d in dfs:
        dsite = d.site
        dtype = d.defect_type
        if dtype == 'Vacancy':
            check,i = is_site_in_structure_coords(dsite,bk,tol=tol)
            el = dsite.specie
            species = Element.from_row_and_group(el.row, el.group+1)
            df.insert(i=i,species=species,coords=dsite.frac_coords)
        elif dtype == 'Interstitial' and sort_to_bulk:
            extra_sites.append(dsite)

    # reorder to match bulk, useful if you don't have the non-relaxed defect structure
    if sort_to_bulk:
        new_structure = sort_sites_to_ref_coords(df, bk, extra_sites,tol=tol)
    # In this case only dummy atoms are inserted, no further changes
    else:
        new_structure = df.copy()
    return new_structure        

        
def defect_finder(structure_defect,structure_bulk,tol=1e-03):
    """
    Function to find defect comparing defect and bulk structure (Pymatgen objects). 

    Parameters
    ----------
    structure_defect : (Pymatgen Structure object)
        Defect structure.
    structure_bulk : (Pymatgen Structure object)
        Bulk structure.
    tol: (float)
        Tolerance for fractional coordinates comparison.

    Returns
    -------
    Defect object
    """
    df = structure_defect
    bk = structure_bulk
    defects = []
    
    for s in df:
        is_site,index = is_site_in_structure_coords(s,bk,tol=tol)
        if is_site:
            if is_site_in_structure(s,bk,tol=tol)[0] == False:
                defect = Substitution(s, bk,site_in_bulk=bk[index])
                defects.append(defect)
        else:
            defect = Interstitial(s, bk)
            defects.append(defect)
            
    for s in bk:
        if is_site_in_structure_coords(s,df,tol=tol)[0] == False:
            defect = Vacancy(s, bk)
            defects.append(defect)
    
    if len(defects) > 1:
        return DefectComplex(defects, structure_bulk)
    elif len(defects) == 1:
        return defects[0]
    else:
        warnings.warn('No defect has been found. Try to adjust the tolerance parameter.',UserWarning)
        return defects



def get_trajectory_for_visualization(structure_defect,structure_bulk,defects=None,tol=1e-03,file=None):
    """
    Create trajectory from defect and bulk structures for visualization in OVITO. 
    The vacancies are shown by inserting in the vacant site the element of same row and next group on the periodic table.

    Parameters
    ----------
    structure_defect : (Pymatgen Structure object)
        Defect structure.
    structure_bulk : (Pymatgen Structure object)
        Bulk structure.
    defects : (tuple or list). 
        Tuple or list of tuples in the format (defect_site,defect_type)
        The format is the same as the output of defect_finder. If None defect_finder is used. The default is None.
    tol: (float)
        Tolerance for site comparison. The distance between sites in target and reference stucture is used, 
        periodicity is accounted for. The tolerance is normalized with respect to lattice vector size. 
        The default is 1e-03.
    file : (str)
        File to save XDATCAR. 
    Returns
    -------
    new_structure (Pymatgen Structure object)
        Structure with dummy atoms as vacancies and interstitials in the bottom.
        The order of the Sites follow the order of the Bulk structure.

    """
    sb = structure_bulk
    dummy = create_def_structure_for_visualization(structure_defect, structure_bulk,defects,sort_to_bulk=True,tol=tol)
    traj = Trajectory.from_structures([dummy,sb],constant_lattice=True)     
    if file:
        if not op.exists(op.dirname(file)):
            os.makedirs(op.dirname(file))
        traj.write_Xdatcar(file)
    return traj
        
        
def write_extxyz_for_visualization(file,structure_defect,structure_bulk,defects=None,tol=1e-03): 
    """
    Write extxyz file for visualization in OVITO. The displacements w.r.t the bulk structure are included.
    The vacancies are shown by inserting in the vacant site the element of same row and next group on the periodic table.
    
    Parameters
    ----------
    file : (str)
        Path to save file. 
    structure_defect : (Pymatgen Structure object)
        Defect structure.
    structure_bulk : (Pymatgen Structure object)
        Bulk structure.
    defects : (tuple or list). 
        Tuple or list of tuples in the format (defect_site,defect_type)
        The format is the same as the output of defect_finder. If None defect_finder is used. The default is None.
    tol: (float)
        Tolerance for site comparison. The distance between sites in target and reference stucture is used, 
        periodicity is accounted for. The tolerance is normalized with respect to lattice vector size. 
        The default is 1e-03.
    """
    sb = structure_bulk
    dummy = create_def_structure_for_visualization(structure_defect, structure_bulk,defects,sort_to_bulk=True,tol=tol)
    write_extxyz_file(file, dummy, sb,displacements=True)
    return


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu May 11 16:33:19 2023

@author: villa
"""
from pynter.testing.core import PynterTest
from pynter.testing.structure import StructureTest


class DefectTest(PynterTest):
    """
    Provides methods to test Defect objects
    """
    def assert_Defect_equal(self,defect1,defect2):
        StructureTest().assert_Site_equal(defect1.site,defect2.site)
        StructureTest().assert_Structure_equal(defect1.bulk_structure,defect2.bulk_structure)
        self.assertEqual(defect1.charge,defect2.charge)
        self.assertEqual(defect1.multiplicity,defect2.multiplicity)
        self.assertEqual(defect1.label,defect2.label)
        
        
class DefectEntryTest(PynterTest):
    """
    Provides methods to test DefectEntry objects
    """
    def assert_DefectEntry_equal(self,entry1,entry2):
        DefectTest().assert_Defect_equal(entry1.defect,entry2.defect)
        StructureTest().assert_Structure_equal(entry1.bulk_structure,entry2.bulk_structure)
        self.assertEqual(entry1.energy_diff,entry2.energy_diff)
        self.assertEqual(entry1.corrections,entry2.corrections)
        self.assertEqual(entry1.data,entry2.data)
        self.assertEqual(entry1.label,entry2.label)
        

            
        
            




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Dec 11 14:33:12 2020

@author: villa
"""

from abc import ABCMeta
from monty.json import MSONable
import numpy as np
import warnings

from pymatgen.core.units import kb

from pynter.defects.structure import defect_finder
from pynter.defects.elasticity import Stresses
from pynter.vasp.utils import get_charge_from_computed_entry



class DefectEntry(MSONable,metaclass=ABCMeta):
    
    def __init__(self,defect,bulk_structure,energy_diff,corrections,data=None,label=None):
        """
        Contains the data for a defect calculation.
        
        Args:
            defect: Defect object (Vacancy, Interstitial, Substitution, Polaron or DefectComplex)
            bulk_structure: Pymatgen Structure without any defects
            energy_diff (float): difference btw energy of defect structure and energy of pure structure
            corrections (dict): Dict of corrections for defect formation energy. All values will be summed and
                                added to the defect formation energy.     
            data : (dict), optional
                Store additional data in dict format.
            label : (str), optional
                Additional label to add to defect specie. Does not influence non equilibrium calculations.
        """
        self._defect = defect
        self._bulk_structure = bulk_structure
        self._energy_diff = energy_diff
        self._corrections = corrections if corrections else {}
        self._data = data if data else {}
        self._defect.set_label(label)
        
    
    def __repr__(self):
        return "DefectEntry: Name=%s, Charge=%i" %(self.name,self.charge)

    def __str__(self):
        output = [
            "DefectEntry",
            "Defect: %s" %(self.defect.__str__()),
            "Bulk System: %s" %self.bulk_structure.composition,
            "Energy: %.4f" %self.energy_diff,
            "Corrections: %.4f" %sum([v for v in self.corrections.values()]),
            "Charge: %i" %self.charge,
            "Multiplicity: %i" %self.multiplicity,
            "Data: %s" %self.data.keys(),
            "Name: %s" %self.name,
            "\n"
            ]
        return "\n".join(output)

    @property
    def bulk_structure(self):
        return self._bulk_structure    

    @property
    def charge(self):
        return self.defect.charge
    
    @property
    def corrections(self):
        return self._corrections
    
    @property
    def data(self):
        """
        Dictionary containing additional data.
        """
        return self._data

    @data.setter
    def data(self,data):
        self._data = data
        return 

    @property
    def defect(self):
        return self._defect  
    
    @property
    def defect_specie(self):
        return self.defect.defect_specie
    
    @property
    def defect_type(self):
        return self.defect.defect_type
    
    @property
    def delta_atoms(self):
        """
        Dictionary with Element as keys and particle difference between defect structure
        and bulk structure as values.
        """
        return self.defect.delta_atoms
    
    @property
    def energy_diff(self):
        return self._energy_diff
    
    @property
    def label(self):
        return self.defect.label
    
    @label.setter
    def label(self,label):
        self.defect.set_label(label)
        return

    @property
    def multiplicity(self):
        return self.defect.multiplicity
    
    @multiplicity.setter
    def multiplicity(self,multiplicity):
        self.defect.set_multiplicity(multiplicity)
        return
            
    @property
    def name(self):
        return self.defect.name
    
    @property
    def symbol(self):
        return self.defect.name.symbol

    @property
    def symbol_charge(self):
        return self.defect.symbol_with_charge

    @property
    def symbol_kroger(self):
        return self.defect.symbol_with_charge_kv
    

    @staticmethod
    def from_computed_entries(computed_entry_defect,computed_entry_bulk,corrections,
                              multiplicity=1,data=None,label=None,tol=1e-03):
        """
        Generate DefectEntry object from Pymatgen's ComputedStructureEntry objects.

        Parameters
        ----------
        computed_entry_defect : (VaspJob)
            ComputedStructureEntry of the defect calculation.
        job_bulk : (VaspJob)
            ComputedStructureEntry of the bulk calculation.
        corrections : (dict)
            Dict of corrections for defect formation energy. All values will be summed and
            added to the defect formation energy.
        multiplicity : (int), optional
            Multiplicity of defect within the supercell. 
            If set to None is attempted to be determined automatically with Pymatgen. The default is 1.
        data : (dict), optional
            Store additional data in dict format.
        label : (str), optional
            Additional label to add to defect specie. Does not influence non equilibrium calculations.
        tol : (float)
            Tolerance for defect_finder function. The default is 1e-03.

        Returns
        -------
        DefectEntry
        """ 
        entry_df, entry_bulk = computed_entry_defect,computed_entry_bulk
        charge = get_charge_from_computed_entry(entry_df)
        energy_diff = entry_df.energy - entry_bulk.energy
        
        return DefectEntry.from_structures(entry_df.structure, entry_bulk.structure, energy_diff,
                                           corrections,charge,multiplicity,data,label,tol=tol)
        

    @staticmethod
    def from_jobs(job_defect, job_bulk, corrections, defect_structure=None,
                  multiplicity=1,data=None,label=None,tol=1e-03):
        """
        Generate DefectEntry object from VaspJob objects.

        Parameters
        ----------
        job_defect : (VaspJob)
            Defect calculation.
        job_bulk : (VaspJob)
            Bulk calculation.
        corrections : (dict)
            Dict of corrections for defect formation energy. All values will be summed and
            added to the defect formation energy.
        defect_structure : (Structure)
            Structure of the defect. If None the intial structure of job_defect is taken. The default is None. 
        multiplicity : (int), optional
            Multiplicity of defect within the supercell. 
            If set to None is attempted to be determined automatically with Pymatgen. The default is 1.
        data : (dict), optional
            Store additional data in dict format.
        label : (str), optional
            Additional label to add to defect specie. Does not influence non equilibrium calculations.
        tol : (float)
            Tolerance for defect_finder function. The default is 1e-03.

        Returns
        -------
        DefectEntry
        """ 
        defect_structure = defect_structure if defect_structure else job_defect.initial_structure
        bulk_structure = job_bulk.final_structure
        energy_diff = job_defect.final_energy - job_bulk.final_energy
        charge = job_defect.charge
        
        return DefectEntry.from_structures(defect_structure, bulk_structure, energy_diff, corrections,
                                                  charge,multiplicity,data,label,tol=tol)


    @staticmethod
    def from_structures(defect_structure,bulk_structure,energy_diff,corrections,charge=0,
                        multiplicity=1,data=None,label=None,tol=1e-03):
        """
        Generate DefectEntry object from Structure objects.

        Parameters
        ----------
        defect_structure : (Structure)
            Defect structure.
        bulk_structure : (Structure)
            Bulk structure.
        energy_diff (float): 
            Difference btw energy of defect structure and energy of pure structure
        corrections (dict): 
            Dict of corrections for defect formation energy. All values will be summed and
            added to the defect formation energy.  
        charge : (int), optional
            Charge of the defect system. The default is 0.
        multiplicity : (int), optional
            multiplicity of defect within the supercell. 
            If set to None is attempted to be determined automatically with Pymatgen. The default is 1.
        data : (dict), optional
            Store additional data in dict format.
        label : (str), optional
            Additional label to add to defect specie. Does not influence non equilibrium calculations.
        tol : (float)
            Tolerance for defect_finder function. The default is 1e-03.

        Returns
        -------
        DefectEntry
        """
        defect = defect_finder(defect_structure, bulk_structure,tol=tol)
        defect.set_charge(charge)
        if multiplicity:
            defect.set_multiplicity(multiplicity)
        else:
            try:
                new_multiplicity = defect.get_multiplicity()
                defect.set_multiplicity(new_multiplicity)
            except NotImplementedError:
                warnings.warn(f'get_multiplicity not implemented for {defect.defect_type}, setting multiplicity to 1')
                defect.set_multiplicity(1)
        
        return DefectEntry(defect, bulk_structure, energy_diff, corrections,data,label)


    def defect_concentration(self, vbm, chemical_potentials, temperature=300, fermi_level=0.0, 
                             per_unit_volume=True,occupation_function='MB'):
        """
        Compute the defect concentration for a temperature and Fermi level.
        Args:
            temperature:
                the temperature in K
            fermi_level:
                the fermi level in eV (with respect to the VBM)
        Returns:
            defects concentration in cm^-3
        """
        n = self.defect.site_concentration_in_cm3 if per_unit_volume else self.multiplicity 
        eform = self.formation_energy(vbm, chemical_potentials, fermi_level=fermi_level)
        
        if occupation_function=='FD':
            conc = n * fermi_dirac(eform,temperature)
        elif occupation_function=='MB':
            conc = n * maxwell_boltzmann(eform,temperature)
        else:
            raise ValueError('Invalid occupation function. Options are: "FD" for Fermi-Dirac and "MB" for Maxwell-Boltzmann.')
        return conc


    def formation_energy(self,vbm,chemical_potentials,fermi_level=0):
        """
        Compute the formation energy for a defect taking into account a given chemical potential and fermi_level
        Args:
            vbm(float): Valence band maximum of pure structure
            chemical_potentials (dict): Dictionary of elemental chemical potential values.
                Keys are Element objects within the defect structure's composition.
                Values are float numbers equal to the atomic chemical potential for that element.
            fermi_level (float):  Value corresponding to the electron chemical potential.
            """
            
        formation_energy = (self.energy_diff + self.charge*(vbm+fermi_level) + 
                       sum([ self.corrections[correction_type]  for correction_type in self.corrections ]) 
                        ) 
        
        if chemical_potentials:
            chempot_correction = -1 * sum([self.delta_atoms[el]*chemical_potentials[el] for el in self.delta_atoms])
        else:
            chempot_correction = 0
            
        formation_energy = formation_energy + chempot_correction
        
        return formation_energy
    
    
    def relaxation_volume(self,stress_bulk,bulk_modulus,add_corrections=True): #still to decide weather to keep this method
        """
        Calculate relaxation volume from stresses. Stresses data needs to be in numpy.array format and present 
        in the "data" dictionary with realtive "stress" key. Duplicate of function that can be found in Stresses
        class in elasticity module, added here for convenience.

        Parameters
        ----------
        stress_bulk : (np.array)
            Stresses of bulk calculation.
        bulk_modulus : (float)
            Bulk modulus in GPa.
        add_corrections : (bool)
            Add correction terms from "elastic_corrections" dict (if key is present in dict).

        Returns
        -------
        rel_volume : (float)
            Relaxation volume in A°^3.
        """
        es = Stresses(stress_bulk)
        return es.get_relaxation_volume(self, bulk_modulus)



def fermi_dirac(E,T):
    """
    Returns the defect occupation as a function of the formation energy,
    using the Fermi-Dirac distribution with chemical potential equal to 0. 
    Args:
        E (float): energy in eV
        T (float): the temperature in kelvin
    """
    return 1. / (1. + np.exp(E/(kb*T)) )


def maxwell_boltzmann(E,T):
    """
    Returns the defect occupation as a function of the formation energy,
    using the exponential dependence of the Maxwell-Boltzmann distribution. 
    This is the more common approach, which is the approximation of the FD-like 
    distribution for N_sites >> N_defects.
    Args:
        E (float): energy in eV
        T (float): the temperature in kelvin
    """
    return np.exp(-1.0*E /(kb*T)) 



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed May 31 16:09:45 2023

@author: villa
"""

import os
import json
from glob import glob
import numpy as np
import warnings
from pprint import pprint

#from pymatgen.electronic_structure.dos import FermiDos # complete dos from_dict bug
from pynter.vasp.pmg.pmg_electronic_structure_dos import CompleteDos

from pynter.data.datasets import Dataset
from pynter.vasp.jobs import VaspJob
from pynter.slurm.job_settings import JobSettings
from pynter.defects.entries import DefectEntry
from pynter.defects.analysis import DefectsAnalysis
from pynter.defects.corrections import get_kumagai_correction_from_jobs
from pynter.phase_diagram.chempots import Chempots, Reservoirs
from pynter.tools.utils import save_object_as_json, get_object_from_json
from pynter.cli import inputs
from pynter.cli.utils import get_dict_from_line_string, round_floats



def setup_defects(subparsers):
    
    parser_defects = subparsers.add_parser('defects',help='Import and analyse defect calculations. Use with extreme care.')
    subparsers_defects = parser_defects.add_subparsers()

    parser_inputs = subparsers_defects.add_parser('inputs',help='Create inputs for VASP DFT calculations')
    setup_inputs(parser_inputs)
    
    parser_import = subparsers_defects.add_parser('import',help='Create defect entries from VASP DFT calculations')
    setup_import(parser_import)
 
    parser_plot = subparsers_defects.add_parser('plot',help='Plot collection of defect entries')
    setup_plot(parser_plot)

    parser_analysis = subparsers_defects.add_parser('analysis',help='Analyse collection of defect entries')
    setup_analysis(parser_analysis)

    return
    


def setup_inputs(parser):
    subparsers = parser.add_subparsers()
    
    parser_vasp = subparsers.add_parser('vasp',help='VASP DFT calculations')
    
    parser_vasp = inputs.parse_common_args(parser_vasp)
    parser_vasp = inputs.parse_vasp_args(parser_vasp)

    parser_vasp.add_argument('-auto','--automation',action='store_true',help='Add default automation to job script',required=False,
                        default=False,dest='automation')    

    parser_vasp.add_argument('-ss','--supercell-size',action='append',help='Size of the supercell (integer or list)',type=int,required=False,
                        default=None,metavar='',dest='supercell_size')

    parser_vasp.add_argument('-rel','--relaxation-scheme',help='Relaxation scheme to use, choose between "default" (2-step PBE) and "gamma" (4-step PBE) (default: %(default)s)',
                        required=False,default='default',type=str,metavar='',dest='relaxation_scheme')
    
    parser_vasp.add_argument('-sub','--substitutions',help=""""Substituions inputs. Provide elements and charges as '{"<new_el>-on-<old_el>":[q0,q1,q2]}'""",type=str,required=False,
                        default=None,metavar='',dest='substitutions')
    
    parser_vasp.add_argument('-vac','--vacancies',help="""Vacancies inputs. Provide elements and charges as '{"el":[q0,q1,q2]}'""",type=str,required=False,
                        default=None,metavar='',dest='vacancies')
    
    parser_vasp.set_defaults(func=create_vasp_inputs)
    return


def create_vasp_inputs(args):
    jobs = []
    schemes = inputs.get_schemes(args)
    if args.substitutions:
        elements_to_replace_with_charges = json.loads(args.substitutions)
        jobs_sub = schemes.substitutions_pbe_relaxation(elements_to_replace_with_charges=elements_to_replace_with_charges,
                                                    supercell_size=args.supercell_size,automation=args.automation,
                                                    locpot=True,rel_scheme=args.relaxation_scheme)
        jobs = jobs + jobs_sub
    if args.vacancies:
        elements_with_charges = json.loads(args.vacancies)
        jobs_vac = schemes.vacancies_pbe_relaxation(elements_with_charges=elements_with_charges,
                                                    supercell_size=args.supercell_size,automation=args.automation,
                                                    locpot=True,rel_scheme=args.relaxation_scheme)
        jobs = jobs + jobs_vac

    ds = Dataset(jobs)
    ds.write_jobs_input()
    return


def setup_import(parser):
    job_script_filename = JobSettings().filename
    parser.add_argument('-pb','--path-bulk',help='Path to bulk calculation',required=True,type=str,metavar='',dest='path_bulk')
    parser.add_argument('-p','--path',help='Path to defect calculations, can contain wildcards (default: %(default)s)',required=False,type=str,default=os.getcwd(),metavar='',dest='path')
    
    parser.add_argument('-e','--exclude',action='append',help='Exclude specific defect types (Vacancy, Substitution, Interstitial, Polaron, DefectComplex)',
                        required=False,default=None,metavar='',dest='exclude')
    parser.add_argument('-c','--corrections',action='store_true',help='Compute Kumagai corrections (default: %(default)s)',required=False,default=False,dest='corrections')
    parser.add_argument('-dt','--dielectric-tensor',help='Dielectric constant or tensor, if tensor write the matrix in a line (a11 a12 a13 a21 a22 a23 a31 a32 a33)',
                        required=False,type=str,default=None,metavar='',dest='dielectric_tensor')    
    parser.add_argument('-j','--job-script-filename',help='Job script filename (default: %(default)s)',required=False,type=str,default=job_script_filename,metavar='',
                        dest='job_script_filename')
    parser.add_argument('-t','--tolerance',help='Tolerance in A° for automatic defect finding (default: %(default)s)',required=False,type=float,default=0.01,metavar='',
                        dest='tolerance')
    parser.add_argument('-s','--savejson',action='store_true',help='Save DefectsAnalysis and bulk DOS objects as json (default: %(default)s)',required=False,default=False,dest='savejson')

    parser.set_defaults(func=import_entries)
    return

def import_entries(args):
    job_bulk = VaspJob.from_directory(args.path_bulk)
    path = args.path
    script = args.job_script_filename
    
    if '*' in path:
        dirs = glob(os.path.normpath(path))
        jobs = [VaspJob.from_directory(path=jdir,job_script_filename=script) for jdir in dirs]
        ds = Dataset(jobs)
    else:
        ds = Dataset.from_directory(path,script)

    def get_last_node(job):
        if len(job.node_points) == 0:
            return None
        else:
            return job.node_points[-1]
    jobs_to_import = []
    for group in ds.groups:
        jobs = ds.select_jobs(groups=[group],is_converged=True)
        charges = []
        for j in jobs:
            if j.charge not in charges:
                charges.append(j.charge)
        for q in charges:
            jobs_same_charge = ds.select_jobs(jobs,charge=q)
            sorted_jobs_same_charge = ds.sort_jobs(jobs_same_charge,features=get_last_node)
            job_defect = sorted_jobs_same_charge[-1]
            if job_defect.formula != job_bulk.formula or job_defect.charge != job_bulk.charge:
                jobs_to_import.append(job_defect)
    print('\nCalculations imported from:')    
    print(ds.jobs_table(jobs_to_import,display=['charge']))
    
    entries = []
    tol = args.tolerance
    for job_defect in jobs_to_import:
        corrections = {}
        if args.corrections:
            if args.dielectric_tensor:
                if len(args.dielectric_tensor.split()) > 1:
                    dielectric_tensor_string = ' '.join(args.dielectric_tensor.split()) # remove extra spaces
                    dielectric_tensor = np.fromstring(dielectric_tensor_string, dtype=float, sep=' ').reshape((3, 3))
            else:
                raise ValueError('Dielectric tensor needs to be provided to perform Kumagai corrections')
            kumagai_corrections = get_kumagai_correction_from_jobs(job_defect, job_bulk, dielectric_tensor,tol=tol)
            kumagai_total = sum([v for v in kumagai_corrections.values()])
            corrections['kumagai'] = kumagai_total
        data = {'stress':job_defect.stress}
        entry = DefectEntry.from_jobs(job_defect,job_bulk,corrections=corrections,data=data,multiplicity=None,tol=tol)
        
        if entry.defect.defect_type == 'DefectComplex' and len(entry.defect.defects) > 3:
            warnings.warn('A complex with more than 3 species has been found, something has likely gone wrong in the automatic defect determination. Excluding it for now...')
        else:    
            entries.append(entry)
    
    band_gap, vbm = job_bulk.energy_gap, job_bulk.vbm
    da = DefectsAnalysis(entries, vbm, band_gap)
    da.filter_entries(inplace=True,exclude=True,types=args.exclude)
    print('\nDefectsAnalysis:')
    print(da.get_dataframe())
    if args.savejson:
        da.to_json('defects_analysis_%s.json' %job_bulk.formula.replace(' ',''))
        job_bulk.get_output_properties(data=['complete_dos'])
        dos = job_bulk.complete_dos
        save_object_as_json(dos,'./DOS_%s.json'%job_bulk.formula.replace(' ',''))
        print('\nDefectsAnalysis object saved as defects_analysis_%s.json' %job_bulk.formula.replace(' ',''))
        print('DOS object saved as DOS_%s.json' %job_bulk.formula.replace(' ',''))
    return
  



def parse_common_args(parser):
    parser.add_argument('-f','--file',help='Json file representation of a DefectAnalysis object',
                        type=str,default=None,metavar='',dest='file')
    
    parser.add_argument('-c','--chempots',help="""Chemical potentials, written in a line ( el1 -x el2 -y el3 -z ). 
                        If not provided, elemental chempots are taken from the Materials Project. Use with care.""",
                        required=False,type=str,default=None,metavar='',dest='chempots')
    
    parser.add_argument('-e','--exclude',action='append',help='Exclude specific defect types (Vacancy, Substitution, Interstitial, Polaron, DefectComplex)',
                        required=False,default=None,metavar='',dest='exclude')
    
    parser.add_argument('-ee','--exclude-elements',action='append',help='Exclude defects containing these elements',required=False,
                    default=None,metavar='',dest='exclude_elements')
    return parser    

def get_defects_analysis(args):
    if args.file:
        da = DefectsAnalysis.from_json(args.file)
        if args.exclude:
            da.filter_entries(inplace=True,exclude=True,types=args.exclude)
        if args.exclude_elements:
            da.filter_entries(inplace=True,exclude=True,elements=args.exclude_elements)
    else:
        raise ValueError('DefectsAnalysis object must be provided as json file')
    return da
        
def get_chempots(args):
    if args.chempots:
        chempots_dict = get_dict_from_line_string(args.chempots)
    else:
        from pynter.tools.materials_project import MPDatabase
        mpd = MPDatabase()
        compositions = args.da.elements
        entries_dict = mpd.get_entries_from_compositions(compositions,lowest_e_above_hull=True)
        chempots_dict = {el:entry.energy_per_atom for el,entry in entries_dict.items()} #get elemental chempots from MP
        warnings.warn('Chemical potentials have not been provided, elemental chemical potentials taken ' \
                      'from the Materials Project database. Use with care.')
        print('Chemical potentials:')
        pprint(round_floats(chempots_dict))
    return Chempots(chempots_dict)
    


def setup_plot(parser):
    parser = parse_common_args(parser)

    parser.add_argument('-b','--binding-energies',action='store_true',help='Plot binding energies',required=False,
                    default=False,dest='plot_binding_energies')

    parser.add_argument('-ctl','--charge-transition-levels',action='store_true',help='Plot formation energies, provide chempots',required=False,
                    default=False,dest='plot_charge_transition_levels')

    parser.add_argument('-eform','--formation-energies',action='store_true',help='Plot formation energies, provide chempots',required=False,
                    default=False,dest='plot_formation_energies')
    
    parser.add_argument('-r','--reservoirs',help='json file with Reservoirs',type=str,required=False,
                    default=None,metavar='',dest='path_reservoirs')

    parser.add_argument('-s','--savefig',action='store_true',help='Save plots as pdf',required=False,
                    default=False,dest='savefig')
    
    parser.set_defaults(func=plot)
    return

def plot(args):
    da = get_defects_analysis(args)
    args.da = da
    if args.plot_binding_energies:
        if len(da.select_entries(types=['DefectComplex'])) == 0:
            raise ValueError('No DefectComplex found. Check that the defect complexes are present in the entries')
        else:
            plt = da.plot_binding_energies()
            if args.savefig:
                plt.savefig('binding_energies.pdf',bbox_inches='tight')
            else:
                plt.show()
    
    if args.plot_charge_transition_levels:
        plt = da.plot_ctl()
        if args.savefig:
            plt.savefig('charge_transition_levels.pdf',bbox_inches='tight')
        else:
            plt.show()
    
    if args.plot_formation_energies:
        if args.path_reservoirs:
            if args.chempots:
                warnings.warn('Reservoirs provided, ignoring input chemical potentials...')
            reservoirs = Reservoirs.from_json(args.path_reservoirs)
            nres = len(reservoirs)
            if nres <= 3:
                nrows,ncolumns = 1, nres
            else:
                ncolumns = 3
                nrows = nres//3 if nres%3==0 else nres//3 +1
            plotsize = (0.83*ncolumns,1.5*nrows)
            index=0
            for r,mu in reservoirs.items():
                index += 1
                plt = da.plot(chemical_potentials=mu,title=r,get_subplot=True,
                        plotsize=plotsize,subplot_settings=[nrows,ncolumns,index])
        else:
            chempots = get_chempots(args)
            plt = da.plot(chempots)
        if args.savefig:
            plt.savefig('formation_energies.pdf',bbox_inches='tight')
        else:
            plt.show()       
    return
    




def setup_analysis(parser):
    parser = parse_common_args(parser)        
    parser.add_argument('-d','--dos',help='Bulk DOS json file',type=str,required=False,default=None,metavar='',dest='dos')
    parser.add_argument('-ef','--fermi',help='Input Fermi level  (default: %(default)s)',required=False,metavar='',default=0,type=float,dest='fermi_level')
    parser.add_argument('-T','--temperature',help='Input temperature  (default: %(default)s)',required=False,default=300,type=float,metavar='',dest='temperature')
    
    parser.add_argument('-cc','--carrier-concentrations',action='store_true',help='Compute carrier concentrations, provide Fermi level, chempots and DOS json file',required=False,
                    default=False,dest='carrier_concentrations')
    
    parser.add_argument('-ctl','--charge-transition-levels',action='store_true',help='Compute charge transition levels',required=False,
                    default=False,dest='charge_transition_levels')
    
    parser.add_argument('-dc','--defect-concentrations',action='store_true',help='Compute defect concentrations, provide Fermi level and chempots',required=False,
                    default=False,dest='defect_concentrations')
    
    parser.add_argument('-eform','--formation-energies',action='store_true',help='Compute formation energies, provide chemical potentials',required=False,
                        default=False,dest='formation_energies')
    
    parser.add_argument('-sef','--solve-fermi-level',action='store_true',help='Solve charge neutrality to get Fermi level, provide chempots and DOS json file',required=False,
                        default=False,dest='solve_fermi_level')
    parser.set_defaults(func=analysis)
    return

def analysis(args):
    da = get_defects_analysis(args)
    args.da = da
    ef = args.fermi_level
    
    chempots = get_chempots(args)
    temperature = args.temperature
    
    if args.charge_transition_levels:
        ctl = da.charge_transition_levels()
        print('\nCharge transition levels:')
        pprint(round_floats(ctl))
    
    if args.formation_energies:
        formation_energies = da.formation_energies(chempots,ef)
        print('\nFormation energies:')
        pprint(round_floats(formation_energies))
    if args.defect_concentrations:
        if args.defect_concentrations:
            print('\nDefect concentrations:')
            print(da.defect_concentrations(chempots,temperature,ef))
        
    dos = get_object_from_json(CompleteDos,args.dos) if args.dos else None
    if args.carrier_concentrations:
        if dos:
            print('\nCarrier concentrations:')
            print(round_floats(da.carrier_concentrations(dos,temperature,ef)))
        else:
            raise ValueError('DOS json file needs to be provided to compute carrier concentrations')
    if args.solve_fermi_level:
        if dos and chempots:
            print('\nEquilibrium Fermi level:\n',round(da.solve_fermi_level(chempots,dos,temperature),4))
        else:
            raise ValueError('DOS json file needs to be provided to compute carrier concentrations')
            
    return
        
    
    
        

        
    



















    
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jun 12 17:52:51 2023

@author: lorenzo
"""
import unittest
from numpy.testing import assert_allclose

from pymatgen.io.vasp.inputs import Kpoints, Poscar, Incar

from pynter.testing.data import JobTest
from pynter.testing.core import PynterTest
from pynter.testing.structure import StructureTest



class VaspInputsTest(PynterTest):
    """
    Provides methods to test pymatgen.io.vasp.inputs objects
    """
    def assert_Incar_equal(self, incar1, incar2, system_only=False):
        """
        Compare Incar objects. Use system_only when default parameters are not known.
        """
        for inc in [incar1,incar2]:
            if type(inc) != Incar:
                inc = Incar(inc)
            for k,v in inc.items():
                inc.__setitem__(k,v)
        if system_only:
            self.assertEqual(incar1['SYSTEM'], incar2['SYSTEM'])
        else:
            assert incar1.diff(incar2)['Different'] == {}

    def assert_Kpoints_equal(self,kpoints1,kpoints2):
        """
        Compare Kpoint objects, compares the strings to avoid inconsistencies
        which can happen when importing/exporting to dict
        """
        kpoints1 = Kpoints.from_string(str(kpoints1))
        kpoints2 = Kpoints.from_string(str(kpoints2))
        assert_allclose(kpoints1.kpts, kpoints2.kpts,atol=1e-03)
        self.assertEqual(kpoints1.style,kpoints2.style)
        self.assertEqual(kpoints1.kpts_shift,kpoints2.kpts_shift)
        self.assertEqual(kpoints1.kpts_weights,kpoints2.kpts_weights)
        
    def assert_Poscar_equal(self,poscar1,poscar2):
        """
        Compare Poscar objects
        """
        poscar1 = Poscar.from_string(str(poscar1))
        poscar2 = Poscar.from_string(str(poscar2))
        self.assertEqual(poscar1.get_string(significant_figures=4),poscar2.get_string(significant_figures=4))
        
    def assert_Potcar_equal(self,potcar1,potcar2):
        """
        Compare Potcar objects
        """
        if len(potcar1) != len(potcar2):
            raise AssertionError
        else:
            for i in range(0,len(potcar1)):
                self.assertEqual(potcar1[i].element, potcar2[i].element)
                self.assertEqual(potcar1[i].functional, potcar2[i].functional)
     
    def assert_VaspInput_equal(self,input1,input2,include_incar=True,system_only=False):
        if include_incar:
            self.assert_Incar_equal(input1['INCAR'],input2['INCAR'],system_only)
        self.assert_Kpoints_equal(input1['KPOINTS'],input2['KPOINTS'])
        self.assert_Poscar_equal(input1['POSCAR'],input2['POSCAR'])
        self.assert_Potcar_equal(input1['POTCAR'],input2['POTCAR'])
        
        
class VaspOutputsTest(PynterTest):
    """
    Provides methods to test pymatgen.io.vasp.outputs objects and pymatgen ComputedEntry
    """
    def assert_ComputedEntry_equal(self,entry1,entry2):
        attributes = ['correction','correction_uncertainty','energy_adjustments',
                      'energy','uncorrected_energy']
        for attr in attributes:
            value1, value2 = getattr(entry1,attr), getattr(entry2,attr)
            self.assert_object_almost_equal(value1,value2)   
        self.assertEqual(entry1.data.keys(),entry2.data.keys())
            
    
    def assert_Vasprun_equal(self,vasprun1,vasprun2):
        self.assert_object_almost_equal(vasprun1.as_dict(),vasprun2.as_dict())
        
        
        
class VaspJobTest(JobTest):
    """
    Provides methods to test VaspJob objects
    """
    def assert_inputs_equal(self,job1,job2,include_incar=True,system_only=False):
        VaspInputsTest().assert_VaspInput_equal(job1.inputs,job2.inputs,include_incar,system_only)

    def assert_outputs_equal(self,job1,job2):
        VaspOutputsTest().assert_ComputedEntry_equal(job1.computed_entry,job2.computed_entry)
        if "Vasprun" in [job1.outputs.keys(),job2.outputs.keys()]:
            VaspOutputsTest().assert_Vasprun_equal(job1.vasprun, job2.vasprun)
            
            
class VaspNEBJobTest(JobTest):
    """
    Provides methods to test VaspNEBJob objects
    """
    def assert_inputs_equal(self,job1,job2,include_incar=True,system_only=False):
        """
        Assert incar, kpoints, potcar and images.
        """
        VaspInputsTest().assert_Incar_equal(job1.incar, job2.incar, system_only=system_only)
        VaspInputsTest().assert_Kpoints_equal(job1.kpoints, job2.kpoints)
        VaspInputsTest().assert_Potcar_equal(job1.potcar, job2.potcar)

        if len(job1.structures) != len(job2.structures):
            raise AssertionError('Number of images differ')
        else:
            for i in range(0,len(job1.structures)):
                StructureTest().assert_Structure_equal(
                    job1.structures[i], job2.structures[i])
            
    def assert_outputs_equal(self,job1,job2):
        """
        Assert pymatgen NEBAnalysis object
        """
        self.assert_object_almost_equal(job1.neb_analysis.as_dict(), job2.neb_analysis.as_dict())
        
        
        
        
        
        